# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.



IM_FROM = 0
IM_FROM_STD = 1
IM_TO = 2
IM_TO_STD = 3
P_FROM = 4
P_FROM_STD = 5
P_TO = 6
P_TO_STD = 7
Q_FROM = 8
Q_FROM_STD = 9
Q_TO = 10
Q_TO_STD = 11
IM_FROM_IDX = 12
P_FROM_IDX = 13
Q_FROM_IDX = 14
IM_TO_IDX = 15
P_TO_IDX = 16
Q_TO_IDX = 17
IA_FROM = 18
IA_FROM_STD = 19
IA_TO = 20
IA_TO_STD = 21
IA_FROM_IDX = 22
IA_TO_IDX = 23


branch_cols_se = 24
