import os
pp_dir = os.path.dirname(os.path.realpath(__file__))

from pandapower._version import __version__, __format_version__
from pandapower.auxiliary import *  # TODO: some functions shouldn't be available with import pandapower
from pandapower.std_types import *
from pandapower.create import *
from pandapower.convert_format import *
from pandapower.file_io import *
from pandapower.sql_io import to_postgresql, from_postgresql, delete_postgresql_net, to_sqlite, from_sqlite
from pandapower.powerflow import *
from pandapower.optimal_powerflow import OPFNotConverged
from pandapower.run import *
from pandapower.toolbox import *  # to be removed -> in future via package namespace available
from pandapower.groups import *
from pandapower.diagnostic import *
from pandapower.runpm import *
from pandapower.pf.runpp_3ph import runpp_3ph

import geojson
geojson.geometry.DEFAULT_PRECISION = 8

import pandas as pd
pd.options.mode.chained_assignment = None  # default='warn'

# import pandapower packages
import pandapower.control
import pandapower.converter
import pandapower.estimation
import pandapower.grid_equivalents
import pandapower.networks
# import pandapower.opf  # no imports available yet
# import pandapower.pf  # no imports available yet
import pandapower.plotting
# import pandapower.protection  # no imports available yet
# import pandapower.pypower  # no imports available yet
import pandapower.shortcircuit
import pandapower.timeseries
import pandapower.toolbox
import pandapower.topology
